{ Komponente jkXML, Version 1.0
  (c) 2000 by Jan W. Krieger, Ismaning, Deutschland
  FREEWARE

  Autor:   Jan W. Krieger
  e-mail:  jan@jkrieger.de
  WWW:     http://www.jkrieger.de/

  Diese Komponente list ein XML-Dokument ein und erzeugt daraus
  eine Baum-Struktur im SPeicher, die dann von einem Programm
  benutzt werden kann, um die Daten und Tags auszulesen.
  Die Komponente beruht auf der W3C-Definitioon fr XML 1.0

  Jedes XML-Dokument wird in einen Prolog-Teil (enthlt Definitionen ...
  und befindet sich vor dem ersten Tag) und einen XML-Teil aufgeteilt.
  Die Node-Typen beider sind gleich.


  -----------------------------
  TjkXML:

  Prolog, XML enthalten je einen Baum von TjkXMLNode (s.u.), die den Inhalt
  von Prolog und Hauptteil einer XML-Datei reprsentieren

  FileName enthlt sen Dateinamen der XML-Datei, die mit Parse XMLFile
  eingelese wird.

  EntityData enthlt eine Liste (TList) mit Pointern auf tjkXMLEntityData-
  Variablen, die die Daten der definierten internen entites (<!ENTITY ...>)
  des XML-Dokumentes enthlt.

  tryInterpret gibt an, ob jkXML versuchen soll in der Procedure ParseXMLFile
  aus der Datei Daten zu ermitteln (interpretieren). Dazu zhlen z.B. Entity-
  Definitionen und bestimmte Daten ber das Dokument, wie encoding, Language ...

  Language enthlt (falls im Dokument definiert) die Sprache des XML-Dokumentes

  Standalone gibt an (falls im Dokument definiert) ob das XML-Dokumentes standalone ist

  Encoding gibt (falls definiert) die Encoding-Angabe des XML-Dokumentes an.

  XMLVersion gibt (falls definiert) die verwendete Version von XML an

  showError gibt an, ob das jkXML Fehlermeldungen anzeigen, oder nur den Fehler-
  Code zurckgeben soll

  geEntityData sucht in EntityData nach dem bergebenen Entity-Namen und gibt die
  Daten zurck

  ParseXMLFile veranlasst die Komponente die in FileName angegebene XML-Datei
  einzulesen.


  -----------------------------
  TjkXMLNode:
  jkXML unterscheidet verschiedene Tags und gibt ihren Typ in NodeType an.

  Der Name eines Tags steht in TagName, sein Inhalt in TagText.
  ParamNames/ParamValues enthalten bei "normalen" tags die Parameternamen und -werte
    Beispiel: <tagname paramname1="paramvalue1" paramname2="paramvalue2">

  Child enthlt eine Liste mit allen Kind.Objekten.

  Parsed gibt an, ob jkXML diesen Tag bereits ausgewertet hat (z.B. bei der
  ?xml-Processing Instruction, oder Entity-Definitionen

  ChildCount gibt die ANzhal der Kinder an

  Level gibt das Level des Objektes im Baum an

  ParamCount gibt die Anzahl von Parametern im Tag an

  ParentNode gibt die bergeordnete Node zurck

  ParsedText enthlt den Inhalt von TagText, nur mit ausgetauschten Entities (interne + predefinied)

  XMLObject enthlt das zugrundeliegende TjkXML-Object

  Mit AddChild/AddChildNum lassen sich Kind-Objekte hinzufgen

  Mit DeleteChild lassen sich Kinder lschen
  -----------------------------

  
  Ismaning: 28. Juli 2000
}

unit jkXML;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, contnrs,
  jktools;

const
  xmlError : Array[0..9] of string =
  ('FEHLER 0: kein Fehler aufgetreten!',
   'FEHLER 1: XML-Tag ist ungltig: SPACES zwischen "<", bzw. "</" und Tag-Name!',
   'FEHLER 2: XML-Tag endet nicht mit ">", oder "["!',
   'FEHLER 3: XML-Tag ist zugleich leer und End-Tag!',
   'FEHLER 4: XML-Tag enthlt "=", ohne vorherigen Parameternamen!',
   'FEHLER 5: Fehler in der "Schachtelstruktur" des XML-Dokumentes!',
   'FEHLER 6: Tag ohne zugehrigen End-Tag gefunden!',
   'FEHLER 7: Tag ist mit sich selber verschachtelt',
   'FEHLER 8: Fehler in Tag-Struktur (ungltige Verschachtelung)',
   'FEHLER 9: XML-Datei enthlt keine Daten');

  xmlSpaces = [#32, #9, #10, #13]; //XML-SPACE-Zeichen
  // Leerzeichen, Tabulator, LineFeed, CarriageReturn

  xmlNames = ['0'..'9', 'a'..'z', 'A'..'Z', '!', '?', ':', '.', '_', '-']; // gltige Zeichen fr XML-Tag-Namen

  xmlValueStart = ['''', '"']; //gltige Zeichen, die einen Parameter-Wert umschlieen


type
  tjkpreXMLRecordType=(xmlpTag, xmlpData);
  pjkpreXMLItem=^tjkpreXMLItem;
  tjkpreXMLItem=record
    ItemType:tjkpreXMLRecordType;
    text: string;
    hasSubordered:boolean;    // nur wichtig fr Prolog-Tags, die andere Tag in [ ] einschlieen
    {z.B.:
      <!DOCTYPE name[

      ]>
    }
  end;

  pjkXMLParameter = ^tjkXMLParameter;
  tjkXMLParameter = record
    name, value:string;
  end;

  tjkXMLNodeType = (xmlRoot, xmlTag, xmlData, xmlComment, xmlPI, xmlDOCTYPE, xmlATTLIST, xmlELEMENT, xmlENTITY, xmlNameDecl, xmlCDATA, xmlNOTATION);
  tjkXMLEntityMode = (emInternal);

  pjkXMLEntityData = ^tjkXMLEntityData;    // Listenobjekt fr die Eigenschaften INTERNER Entity-Definitionen
  tjkXMLEntityData = record                // externe Entities werden (noch) nicht untersttzt
    name, data:string;        // der Entity und ihr Inhalt
    mode:tjkXMLEntityMode;    // bis jetzt nur wmInternal, fr Erweiterung um externe Entities vorgesehen
  end;

  TjkXMLNode = class(TPersistent)   // Baumknoten-Objekt fr XML-Baum (DOM)
  private
    FItem:TObjectList;    // Enthlt eine Liste mit allen Kind-Objekten
    FParent:TjkXMLNode;   // enthlt das bergeordnete Node, oder nil, wenn es das erste Node des Baumes ist
    FLevel:longint;       // entht das Level des Nodes. Es ist immer um eins hher, als das Level
                          // des Vaters, oder 0 fr das erste Element des Baumes

    function getItem(i:longint):TjkXMLNode;
    procedure SetItem(i:longint; value:TjkXMLNode);
    function GetChildCount:longint;
    function GetParamCount:longint;
    function getParsedText: string;   // tauscht Entities gegen ihre entspr. Daten aus
  public
    TagText, TagName:string;
    Nodetype: tjkXMLNodeType;
    Parsed: boolean;      // gibt an, ob die XML-Komponente bereits Daten aus diesem Tag gezogen hat
                          // z.B. <?xml ... ?>, bestimmte <!DOCTYPE ... > usw.
    ParamNames:TStrings;  // Zugriff auf alle Parameter-Namen des Nodes
    ParamValues:TStrings;  // Zugriff auf alle Parameter-Werte des Nodes

    XMLObject:TComponent;

    property Child[i:longint]:TjkXMLNode read getItem write setitem; default;  // Zugriff auf alle Kind-Nodes des Nodes

    constructor create(AOwner: TjkXMLNode; XMLObj:TComponent);
    destructor destroy; override;
    function getParsedString(txt:string):string; // parsed alle Entities aus txt heraus (durch ersetzen)
    function AddChild:TjkXMLNode;  // fgt ein neues Kind hinzu und gibt die Objekt-Referenz zurck
    function AddChildNum:longint;  // fgt ein neues Kind hinzu und gibt die Nummer des neuen Objekts in Child[i] zurck
    function DeleteChild(node:TjkXMLNode):boolean; overload; // Lscht das angegebene Node aus der Liste, gibt bei Erfolg true zurck
    function DeleteChild(node:longint):boolean; overload;    // Lscht das angegebene Node aus der Liste, gibt bei Erfolg true zurck
    procedure clear;                                         // lscht alle Kind-Elemente und setzt die Tag-Eigenschaften zurck
    function hasChildren: boolean;                           // gibt an, ob Kind-Nodes vorhanden sind
    function getFirstChildNode(ntype: TjkxmlNodeType): TjkXMLNode; overload; //gibt das erste Kind vom Typ ntype zurck
    function getFirstChildNode(nam: string): TjkxmlNode; overload;           //gibt das erste Kind mit dem Namen nam zurck
    function getParamValue(nam: string): string;           
    function getParamValueParsed(nam: string): string;
    function ParamExists(nam: string): boolean;
  published
    property ChildCount:Longint read getChildCount; // gibt die Anzahl der Kind-Nodes zurck
    property ParamCount: longint read getParamCount;
    property Level:Longint read FLevel;             // gibt das Level des Nodes an
    property ParentNode:TjkXMLNode read FParent write FParent; // gibt das ParentNode an, oder nil, wenn es sich um das unterste Node im Baum handelt
    property ParsedText:string read getParsedText;  // TagText, mit ausgetauschten entities
  end;

  TjkXML = class(TComponent)
  private
    { Private-Deklarationen }
    preXMLList: TList;
    FFileName:TFileName;
    Ftags:TStrings;
    FShowError, FtryInt:boolean;
    FLanguage:string;
    FXMLVer:string;
    FEncoding:string;
    FSTandalone:boolean;
    FConvertSpaces:boolean;

//    function Interpret1Tag(tagtext:string; var tagname:string; var isendtag, isemptytag:boolean; var tagParams:TStrings; var TagType:tjkXMLNodeType):word;
    // lst einen XML-Tag in seine "Elemente" auf --> tagname, Parameter, handelt es sich um einen Abschlusstag

    function InterpretTag(tagIn:string; var tagname, tagtext:string; var isendtag, isemptytag:boolean; var tagParams:TStrings; var TagType:tjkXMLNodeType):word;
    // lst einen XML-Tag in seine "Elemente" auf --> tagname, Parameter, handelt es sich um einen Abschlusstag

    function InterpretPI_xml(pi:string):longint;
    function InterpretEntityDef(entity:string):longint;
    function convertXMLText(text: string): string; // konvertiert den Parameter text folgendermaen:
                                                   //     - mehrfache Leerzeichen (#9, #10, #13, #32) werden durch einfache SPACEs (#32) ersetzt
                                                   //     - fhrende und abschlieende Leerzeichen werden entfernt
    {function getEntity(i:longint):tjkXMLEntityData;
    procedure setEntity(i:longint; value:tjkXMLEntityData);}
  protected
    { Protected-Deklarationen }
  public
    { Public-Deklarationen }
    testOutput:TStrings;
    Prolog, XML : TjkXMLNode;   // Fupunkte der Bume fr Prolog und eigentlichen XML-Quelltext
    EntityData:TList;           // Liste mit tjkXMLEntityData-Objekten fr INTERNE Entity-Definitionen

    //property EntityData[i:longint]:tjkXMLEntityData read getEntity; write setEntity;

    constructor create(AOwner: TComponent); override;
    destructor destroy; override;

    function ParseXMLFile:word;  // liest die XML-Datei (FileName) ein
    function getEntityData(entity:string):string;    // sucht in EntityData eine Entity mit dem Namen entity und gibt deren data zurck; ansonsten ''
    function ParseEntityString(txt: string): string; // parsed einen String, der Entities enthlt
  published
    { Published-Deklarationen }
    property FileName:TFileName read FFileName write FFileName;  // gibt die einzulesende XML-Datei an
    property showError:boolean read fShowError write FShowError; // gibt an, ob Fehlermeldungen angezeigt werden sollen.
    property tryInterpret:boolean read ftryint write ftryint;    // gibt an, ob der XML-parser versuchen soll, ihm bekannte XML-tags auszuwerten.
                                                                 // bei diesen wird im Objekt tjkXMLNode die Variable parsed auf true gesetzt
    property XMLVersion:string read FXMLver write FXMLVer;       // enthlt die Versionsnummer der im Dokument verwendeten XML-Version (falls angegeben)
    property Language:string read FLanguage write FLanguage;     // falls Definiert: Sprache des XML-Dokumentes
    property Encoding:string read FEncoding write FEncoding;     // falls Definiert: Encoding des XML-Dokumentes
    property Standalone:boolean read FStandalone write FStandalone;     // falls Definiert: gibt an, ob das XML-Dokument 'standalone' ist. (keine berprfung, nur auswertung von <?xml ... ?>
    property ConvertSpaces:boolean read FConvertSpaces write FConvertSpaces;  // falls true: alle mehrfachen Leerraumzeichen werden gegen ein Leerzeichen ausgetauscht, fhrende Lehrzeichen werden entfernt 
  end;

procedure Register;
function EntityString(txt: string): string;  // ersetzt diverse Zeichen (siehe bei Prozedure) durch entsprechende Entities

implementation

procedure Register;
begin
  RegisterComponents('JKSoft', [TjkXML]);
end;


constructor TjkXMLNode.create(AOwner: TjkXMLNode; XMLObj:TComponent);
begin
  if assigned(AOwner) then begin
    FParent:=AOwner;          // Der Eigner eines jeden Objektes ist sein Vater
    FLevel:=FParent.Level+1;  // Das Level eines neuen Nodes ist immer um eins hher, als das des Vaters
  end else begin
    FParent:=nil;      // Das erste Element des Baumes hat keinen Vater (FParent:=NIL)
    FLevel:=0;         // es hat das Level 0
  end;
  Fitem:=TObjectList.create;
  FItem.ownsobjects:=true;
  ParamNames:= TStringList.create;
  ParamValues:= TStringList.create;
  TagText:='';
  TagName:='';
  Nodetype:=xmlData;
  XMLObject:=XMLObj;
  Parsed:=false;
end;

destructor TjkXMLNode.destroy;
begin
  FItem.clear;
  FItem.free;
  ParamNames.free;
  ParamValues.free;
  inherited destroy;
end;

procedure TjkXMLNode.clear;
begin
  FItem.clear;
  NodeType:=xmlData;
  ParamNames.clear;
  ParamValues.clear;
  TagName:='';
  TagText:='';
  FLevel:=0;
end;

function TjkXMLNode.getParsedString(txt:string):string;
var i:longint;
    isEntity, isPEntity :boolean;
    res, tmp:string;
    r:longint;
begin
  result:=TjkXML(XMLObject).ParseEntityString(Txt);
  {if (NodeType in [xmlData])and (length(txt)>2) then begin
    isEntity:=false;
    isPEntity:=false;
    res:='';
    tmp:='';
    for i:=1 to length(txt) do begin
      if txt[i]='&' then begin isEntity:=true; tmp:=''; end;
      //if txt[i]='%' then begin isPEntity:=true; tmp:=''; end;
      if not isEntity then res:=res+txt[i];
      if isEntity and (not(txt[i] in [';', '&', '%', ' ', #10, #13, #9])) then tmp:=tmp+txt[i];
      //if isPEntity and (not(txt[i] in [';', '&', '%', ' ', #10, #13, #9])) then tmp:=tmp+txt[i];
      //if isPEntity and (txt[i]=';') then begin isPEntity:=false; tmp:=''; end;
      if isEntity and (txt[i]=';') then begin
        isEntity:=false;
        if length(tmp)>0 then
          if copy(tmp,1,2)='#x' then try r:=hextodec(copy(tmp,3,length(tmp)-2)); res:=res+chr(r) except end
            else if tmp[1]='#' then try r:=strtoint(copy(tmp,2,length(tmp)-1)); res:=res+chr(r) except end
              else if tmp='apos' then res:=res+''''
                else if tmp='quot' then res:=res+'"'
                  else if tmp='amp' then res:=res+'&'
                    else if tmp='lt' then res:=res+'<'
                      else if tmp='gt' then res:=res+'>'
                        else if tmp='nbsp' then res:=res+' '//entspricht nicht ganz der XML-Definition !!!
                          else res:=res+TjkXML(XMLObject).getEntityData(tmp);

        tmp:='';
      end;
    end;
    //if isEntity then res:=res+tmp;
    result:=res;
  end; }
end;

function TjkXMLNode.getItem(i:longint):TjkXMLNode;
begin
  result:=TjkXMLNode(FItem[i]);
end;

procedure TjkXMLNode.SetItem(i:longint; value:TjkXMLNode);
begin
  FItem[i]:=value;
end;

function TjkXMLNode.GetChildCount:longint;
begin
  result:= FItem.Count;
end;

function TjkXMLNode.GetparamCount:longint;
begin
  result:= ParamNames.Count;
end;

function TjkXMLNode.AddChild:TjkXMLNode;
var node:TjkXMLNode;
begin
  node:=TJKXMLNode.create(self, XMLObject);
  FItem.Add(node);
  result:=node;
end;

function TjkXMLNode.AddChildNum:longint;
var node:TjkXMLNode;
begin
  node:=TJKXMLNode.create(self, XMLObject);
  result:=FItem.Add(node);
end;

function TjkXMLNode.DeleteChild(node:TjkXMLNode):boolean;
var i:longint;
begin
  result:=false;
  if assigned(node) then begin
    i:=FItem.IndexOf(node);
    if i>-1 then begin
      node.destroy;
      FItem.Delete(i);
      result:=true;
    end;
  end;
end;

function TjkXMLNode.DeleteChild(node:longint):boolean;
begin
  result:=false;
  if node in [0.. FItem.count-1] then begin
    //tJKXMLNode(FItem[node]).clear;
    tJKXMLNode(FItem[node]).destroy;
    FItem.Delete(node);
    result:=true;
  end;
end;







constructor TjkXML.create(AOwner: TComponent);
begin
  inherited create(AOwner);
  preXMLList:=TList.Create;
  EntityData:=TList.Create;
  testOutput:=TStringList.create;
  FTags:=TStringList.create;
  FTags.clear;
  Prolog := TjkXMLNode.create(nil, self);
  Prolog.Nodetype:=xmlRoot;
  XML := TjkXMLNode.create(nil, self);
  XML.Nodetype:=xmlRoot;
  FShowError:=true;
  FtryInt:=true;
  FLanguage:='';
  FXMLVer:='';
  FEncoding:='';
  FStandalone:=false;
  FConvertSpaces:=true;
end;

destructor TjkXML.destroy;
var i:longint;
begin
  if preXMLList.Count>0 then
    for i:=preXMLList.Count-1 downto 0 do
      dispose(preXMLList.items[i]);
  preXMLList.clear;
  preXMLList.free;
  EntityData.clear;
  testOutput.free;
  Prolog.destroy;
  FTags.free;
  xml.destroy;
  inherited destroy;
end;


{function TjkXML.getEntity(i:longint):tjkXMLEntityData;
begin
  result:=tjkXMLEntityData(FEntities[i]^)
end;

procedure TjkXML.setEntity(i:longint; value:tjkXMLEntityData);
begin
  tjkXMLEntityData(FEntities[i]^):=value;
end;}

function TjkXML.getEntityData(entity:string):string;
var i:longint;
begin
  if EntityData.count>0 then for i:=0 to EntityData.Count-1 do
    if tjkXMLEntityData(EntityData[i]^).name=entity then result:=tjkXMLEntityData(EntityData[i]^).data;
end;

function TjkXML.InterpretPI_xml(pi:string):longint;
var ValueStart, tag, tagname, ParamName, ParamValue:string;
    i, taglen:longint;
    isTagName, isParamName, isParamValue:boolean;
    errorcode:longint;
begin
  result:=0;
  tag:=copy(pi,3,length(pi)-4);
  tag:=delRSpace(tag);
  taglen:=length(tag);
  tagname:='';
  ParamName:='';
  ParamValue:='';
  ValueStart:='';

    istagName:=true; // zuerst nach Tagnamen suchen
    i:=1;
    repeat
      if (isTagName and (tag[i]='=')) then errorcode:=4;
      if isTagName and (tag[i] in xmlNames) then tagname:=tagName+tag[i];
      if isTagName and (not(tag[i] in xmlNames)) then begin
        isTagName:=false;
        ParamName:='';
        repeat   // nach dem Anfang des Parameternamens suchen
          inc(i);
        until (tag[i] in xmlnames)or(i=taglen);
        if i<=taglen then begin isParamName:=true; end;
      end;
      if isParamName and (tag[i] in xmlNames) then paramName:=paramName+tag[i];
      if isParamName and (tag[i] in ['=']+xmlSpaces) then begin
        isParamName:=false;
        ParamValue:='';
        repeat   // nach dem Anfang des Parameterwertes suchen
          inc(i);
        until (tag[i] in xmlValueStart)or(i=taglen);
        if i<=taglen then begin isParamValue:=true; ValueStart:=tag[i]; inc(i); end;
      end;
      if isParamValue and (not(tag[i] = ValueStart)) then ParamValue:=ParamValue+tag[i];
      if isParamValue and (tag[i] = ValueStart) then begin
        isParamValue:=false;
        // ParamName & ParamValue auswerten:
        //ShowMessage(ParamName+#13#10+ParamValue);
        if ParamName='encoding' then FEncoding:=ParamValue
          else if ParamName='version' then FXMLVer:=ParamValue
            else if ParamName='standalone' then FStandAlone:=(ParamValue='yes');

        // nach nchstem ParamName suchen:
        ParamName:='';
        repeat   // nach dem Anfang des Parameternamens suchen
          inc(i);
        until (tag[i] in xmlnames)or(i=taglen);
        if i<=taglen then begin isParamName:=true; dec(i) end;
      end;
      inc(i);
    until (i>taglen)or(errorcode>0);
  result:=errorcode;
end;

function TjkXML.InterpretEntityDef(entity:string):longint;
var tag:string;
    i:longint;
    isName, isData:boolean;
    nam, dat:string;
    data:pjkxmlEntityData;
begin
  result:=0;
  isName:=true;
  isData:=false;
  nam:='';
  dat:='';
  tag:=copy(entity, 10, length(entity)-9-1);
  for i:=1 to length(tag) do begin
    if isName then if tag[i] in xmlNames then nam:=nam+tag[i];
    if isName and (tag[i] in xmlValueStart) then begin isName:=false; isData:=true; end;
    if isData and (not(tag[i] in xmlValueStart)) then dat:=dat+tag[i];
  end;
  if isName and(not isData) then result:=-1;
  if isData and(not isName) then begin
    new(data);
    data^.name:=nam;
    data^.data:=dat;
    data^.mode:= emINTERNAL;
    EntityData.Add(data);
  end;
end;


function TjkXML.InterpretTag(tagIn:string; var tagname, tagtext:string; var isendtag, isemptytag:boolean; var tagParams:TStrings; var TagType:tjkXMLNodeType):word;
var tag, ValueStartChar, paramTmp, valueTmp, ParamText, ValueStart : ansistring;
    isParamName, isParamValue, isTagName : boolean;
    i, taglen:longint;
    errorcode:Word;
begin
  //Initialisierung aller Parameter
  tagParams.Clear;
  tagname           := '';
  tagText           := '';
  TagType           := xmlTag;
  isEndTag          := false;
  isEmptyTag        := false;
  tag               := delrspace(tagIn);
  taglen            := length(tag);
  isParamName       := false;
  isParamValue      := false;
  errorcode         := 0;
  ValueStartChar    := '';
  paramTmp          := '';
  valueTmp          := '';
  ParamText         := '';
  ValueStart        := '"';

  // Suche nach einleitendem "<", oder "</"  --> Bestimmung, ob Endtag;
  // Suche nach ausleitendem "/>"  --> Bestimmung leerer Tag;
  // Bestimmen weiterer formaler Fehler
  if (tag[1]+tag[2]='</') then begin
    isEndTag:=true;  // der wird durch "</" eingeleitet
                     // ==> es handelt sich um Endtag
  end;
  if (tag[taglen-1]+tag[taglen]='/>') then begin
    isEmptyTag:=true;  // der Tag wird durch "/>" abgeschlossen
                       // ==> es handelt sich um einen leeren Tag
  end;
  if isEndTag and isEmptyTag then errorcode:=3;
  if (tag[taglen]<>'>')and(tag[taglen]<>'[') then errorcode:=2;

  // wenn es sich auch um einen "nicht-normalen" Tag handeln kann, wird dieser spezifiziert
  if (not isEndTag) and (not isEmptyTag) then
    if (copy(tag,1,4)='<!--') and (copy(tag,taglen-2,3)='-->') then TagType:=xmlComment
      else if (copy(tag,1,2)='<?') and (copy(tag,taglen-1,2)='?>') then TagType:=xmlPI
        else if (copy(tag,1,9)='<![CDATA[') then TagType:=xmlCDATA
          else if (copy(tag,1,9)='<!DOCTYPE') then TagType:=xmlDOCTYPE
            else if (copy(tag,1,9)='<!ATTLIST') then TagType:=xmlATTLIST
              else if (copy(tag,1,8)='<!ENTITY') then TagType:=xmlENTITY
                else if (copy(tag,1,9)='<!ELEMENT') then TagType:=xmlELEMENT
                  else if (copy(tag,1,4)='<![%') then TagType:=xmlNameDecl
                    else if (copy(tag,1,10)='<!NOTATION') then TagType:=xmlNOTATION;


  // xmlTag, xmlData, xmlComment, xmlPI, xmlDOCTYPE, xmlATTLIST, xmlELEMENT, xmlENTITY, xmlNameDecl, xmlCDATA
  // Falls keine Fehler aufgetreten sind verarbeiten des Standard-Tags
  if (errorcode=0)and(TagType=xmlTag) then begin
    // Tagumklammerung entfernen
    if isEndTag then tag:=copy(tag,3,taglen-2) else tag:=copy(tag,2,taglen-1);
    taglen:=length(tag);
    if isEmptyTag then tag:=copy(tag,1,taglen-2) else tag:=copy(tag,1,taglen-1);
    tag:=delRSpace(tag);
    taglen:=length(tag);
    //tagParams.Add(tag);  // Testausgabe

    // Standard-Tags interpretieren
    istagName:=true; // zuerst nach Tagnamen suchen
    i:=1;
    repeat
      if (isTagName and (tag[i]='=')) then errorcode:=4;
      if isTagName and (tag[i] in xmlNames) then tagname:=tagName+tag[i];
      if isTagName and (not(tag[i] in xmlNames)) then begin
        isTagName:=false;
        ParamTmp:='';
        if i<taglen then repeat   // nach dem Anfang des Parameternamens suchen
          inc(i);
        until (tag[i] in xmlnames)or(i=taglen);
        if i<taglen then begin isParamName:=true; end;
      end;
      if isParamName and (tag[i] in xmlNames) then paramtmp:=paramtmp+tag[i];
      if isParamName and (tag[i] in ['=']+xmlSpaces) then begin
        isParamName:=false;
        tagParams.add(Paramtmp);
        ParamTmp:='';
        if i<taglen then repeat   // nach dem Anfang des Parameterwertes suchen
          inc(i);
        until (tag[i] in xmlValueStart)or(i=taglen);
        if i<taglen then begin isParamValue:=true; ValueStart:=tag[i]; inc(i); end;
      end;
      if isParamValue and (not(tag[i] = ValueStart)) then valuetmp:=valuetmp+tag[i];
      if isParamValue and (tag[i] = ValueStart) then begin
        isParamValue:=false;
        tagParams.Add(valuetmp);
        valuetmp:='';
        if i<taglen then repeat   // nach dem Anfang des Parameternamens suchen
          inc(i);
        until (i=taglen)or(tag[i] in xmlnames);
        if i<taglen then begin isParamName:=true; dec(i) end;
      end;
      inc(i);
    until i>taglen;
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten des Kommentar-Tags
  if (errorcode=0)and(TagType=xmlComment) then begin
    // Tagumklammerung entfernen
    tagtext:=copy(tag,5,taglen-7)
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten des CDATA-Tags
  if (errorcode=0)and(TagType=xmlCDATA) then begin
    // Tagumklammerung entfernen
    tagtext:=copy(tag,10,taglen-9-3)
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten des NOTATION-Tags
  if (errorcode=0)and(TagType=xmlNOTATION) then begin
    // Tagumklammerung entfernen
    tagtext:=copy(tag,11,taglen-10-1)
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten einer Processing Instruction
  if (errorcode=0)and(TagType=xmlPI) then begin
    // Tagumklammerung entfernen
    tag:=copy(tag,3,taglen-4);
    tag:=delRSpace(tag);
    taglen:=length(tag);
    //tagParams.Add(tag);  // Testausgabe

    istagName:=true; // zuerst nach Tagnamen suchen
    i:=1;
    repeat
      if (isTagName and (tag[i]='=')) then errorcode:=4;
      if isTagName and (tag[i] in xmlNames) then tagname:=tagName+tag[i];
      if isTagName and (not(tag[i] in xmlNames)) then begin
        isTagName:=false;
        ParamTmp:='';
        if i<taglen then repeat   // nach dem Anfang des Parameternamens suchen
          inc(i);
        until (tag[i] in xmlnames)or(i=taglen);
        if i<taglen then begin isParamName:=true; end;
      end;
      if isParamName and (tag[i] in xmlNames) then paramtmp:=paramtmp+tag[i];
      if isParamName and (tag[i] in ['=']+xmlSpaces) then begin
        isParamName:=false;
        tagParams.add(Paramtmp);
        ParamTmp:='';
        if i<taglen then repeat   // nach dem Anfang des Parameterwertes suchen
          inc(i);
        until (tag[i] in xmlValueStart)or(i=taglen);
        if i<taglen then begin isParamValue:=true; ValueStart:=tag[i]; inc(i); end;
      end;
      if isParamValue and (not(tag[i] = ValueStart)) then valuetmp:=valuetmp+tag[i];
      if isParamValue and (tag[i] = ValueStart) then begin
        isParamValue:=false;
        tagParams.Add(valuetmp);
        valuetmp:='';
        if i<taglen then repeat   // nach dem Anfang des Parameternamens suchen
          inc(i);
        until (tag[i] in xmlnames)or(i=taglen);
        if i<taglen then begin isParamName:=true; dec(i) end;
      end;
      inc(i);
    until i>taglen;
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten einer DOCTYPE-Deklaration
  if (errorcode=0)and(TagType=xmlDOCTYPE) then begin
    // Tagumklammerung entfernen
    if tag[length(tag)]='>' then begin
      tagtext:=copy(tag,10,taglen-10); // gltig fr <!DOCTYPE ohne [...]>
    end else begin
      tagtext:='';
      tagname:=copy(tag,10,taglen-10);
      //Falls Tag mit ]> beendet --> text anders ermitteln
    end;
  end;

  // xmlELEMENT, xmlENTITY, xmlNameDecl
  // Falls keine Fehler aufgetreten sind verarbeiten einer Processing Instruction
  if (errorcode=0)and(TagType=xmlATTLIST) then begin
    // Tagumklammerung entfernen
    tagtext:=copy(tag,10,taglen-10)
  end;

  // xmlENTITY, xmlNameDecl
  // Falls keine Fehler aufgetreten sind verarbeiten einer Processing Instruction
  if (errorcode=0)and(TagType=xmlELEMENT) then begin
    // Tagumklammerung entfernen
    tagtext:=copy(tag,10,taglen-10)
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten einer Processing Instruction
  if (errorcode=0)and(TagType=xmlENTITY) then begin
    // Tagumklammerung entfernen
    tagtext:=copy(tag,9,taglen-9)
  end;

  // Falls keine Fehler aufgetreten sind verarbeiten einer Processing Instruction
  if (errorcode=0)and(TagType=xmlNameDecl) then begin
    tagname:=copy(tag,5,taglen-6);
  end;




//  tagParams.Add(#13#10+xmlError[errorcode]);  //Testausgabe des Fehler-Statuses
  result:=errorcode;
end;

function TjkXML.ParseXMLFile:word;
var f:file of char;
    data:pjkpreXMLItem;
    tmp:char;
    isend, isempty, isTag, isValue, isText, isProlog:boolean;
    err, i, j:longint;
    tagname, tagtext, txt:string;
    prnt, node : TjkXMLNode;
    sl:TStrings;
    tagtype:tjkXMLNodeType;
begin
  { zuerst wird die XML-Datei in eine Liste eingelese und zwar ist jeder
    Tag- und jeder Text-Abschnitt je ein Eintrag
    Beispiel:
      <CDs><CD>name1</CD><CD>name 2</CD></CDs>
      wird zu:
        <CDs>
        <CD>
        name1
        </CD>
        <CD>
        name 2
        </CD>
        </CDs>                                                           }

  if preXMLList.Count>0 then     // Listen-Objekt leeren
    for i:=preXMLList.Count-1 downto 0 do
      dispose(preXMLList.items[i]);
  preXMLList.clear;
  sl:=TStringList.create;
  sl.clear;
  assignfile(f,FFileName);
  Reset(f);
  result:=0;
  if (filesize(f)=0) and FShowError then begin
    ShowMessage(xmlerror[9]);
    result:=9;
    exit;
  end;
  new(data);
  preXMLList.Add(data);
  Read(f,tmp);
  if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
  data^.text:=tmp;
  data^.hasSubordered:=false;
  isTag:=false;
  isValue:=false;
  isText:=false;
  if tmp='<' then isTag:=true else isText:=true;
  while not eof(f) do begin
    Read(f,tmp);

    if isTag and ((tmp<>'>')or isValue) then data^.text:=data^.text+tmp;
    if isText and (tmp<>'<') then data^.text:=data^.text+tmp;
    if isTag and((tmp ='''')or(tmp='"')) then isValue:=not isValue;
    if isTag and(not isValue) and (tmp='[') and ((copy(data^.text,1,4)='<![%')or(copy(data^.text,1,9)='<!DOCTYPE')) then begin
//      data^.text:=data^.text + tmp;
        data^.text:=stringreplace(data^.text, #13, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.text:=stringreplace(data^.text, #10#10, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.hasSubordered:=true;
     //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
        if not eof(f) then begin
          new(data);
          preXMLList.Add(data);
          Read(f,tmp);
          if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
          data^.text:=tmp;
          data^.hasSubordered:=false;
          isTag:=false;
          isValue:=false;
          isText:=false;
          if tmp='<' then isTag:=true else isText:=true;
        end;
      end;
    if isTag and(tmp='[') and (data^.text='<![CDATA[') then begin
    //oder  if isTag and (tmp='[') and ((data^.text='<![CDATA[')or(copy(data^.text,1,4)='<![%')) then begin
        while (not eof(f))and(copy(data^.text,length(data^.text)-2,3)<>']]>') do begin
          Read(f,tmp);
          data^.text:=data^.text+tmp;
        end;
        data^.text:=stringreplace(data^.text, #13, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.text:=stringreplace(data^.text, #10#10, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.hasSubordered:=false;
        //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
        if not eof(f) then begin
          new(data);
          preXMLList.Add(data);
          Read(f,tmp);
          if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
          data^.text:=tmp;
          data^.hasSubordered:=false;
          isTag:=false;
          isValue:=false;
          isText:=false;
          if tmp='<' then isTag:=true else isText:=true;
        end;
      end;
    if isTag and(data^.text='<!--') then begin
        while (not eof(f))and(copy(data^.text,length(data^.text)-2,3)<>'-->') do begin
          Read(f,tmp);
          data^.text:=data^.text+tmp;
        end;
        data^.text:=stringreplace(data^.text, #13, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.text:=stringreplace(data^.text, #10#10, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.hasSubordered:=false;
        //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
        if not eof(f) then begin
          new(data);
          preXMLList.Add(data);
          Read(f,tmp);
          if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
          data^.text:=tmp;
          data^.hasSubordered:=false;
          isTag:=false;
          isValue:=false;
          isText:=false;
          if tmp='<' then isTag:=true else isText:=true;
        end;
      end;
    if isText and(copy(data^.text,length(data^.text)-2,3)=']]>') then begin
        data.text:=copy(data^.text,1,length(data^.text)-3);
        data^.text:=stringreplace(data^.text, #13, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.text:=stringreplace(data^.text, #10#10, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.hasSubordered:=false;
        //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
        new(data);
        preXMLList.Add(data);
        data^.ItemType:=xmlpdata;
        data^.text:=']]>';
        if not eof(f) then begin
          new(data);
          preXMLList.Add(data);
          Read(f,tmp);
          if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
          data^.text:=tmp;
          data^.hasSubordered:=false;
          isTag:=false;
          isValue:=false;
          isText:=false;
          if tmp='<' then isTag:=true else isText:=true;
        end;
      end;
    if isText and(copy(data^.text,length(data^.text)-1,2)=']>') then begin
        data.text:=copy(data^.text,1,length(data^.text)-2);
        data^.text:=stringreplace(data^.text, #13, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.text:=stringreplace(data^.text, #10#10, #10, [rfReplaceAll, rfIgnoreCase]);
        data^.hasSubordered:=false;
        //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
        new(data);
        preXMLList.Add(data);
        data^.ItemType:=xmlpData;
        data^.text:=']>';
        if not eof(f) then begin
          new(data);
          preXMLList.Add(data);
          Read(f,tmp);
          if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
          data^.text:=tmp;
          data^.hasSubordered:=false;
          isTag:=false;
          isValue:=false;
          isText:=false;
          if tmp='<' then isTag:=true else isText:=true;
        end;
      end;
    if isTag and (not isValue) and (tmp='>') then begin
      data^.text:=data^.text + tmp;
      data^.text:=stringreplace(data^.text, #13, ' ', [rfReplaceAll, rfIgnoreCase]);
      data^.text:=stringreplace(data^.text, #10, ' ', [rfReplaceAll, rfIgnoreCase]);
      data^.hasSubordered:=false;
      //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
      if not eof(f) then begin
        new(data);
        preXMLList.Add(data);
        Read(f,tmp);
        if tmp='<' then data^.ItemType:=xmlpTag else data^.ItemType:=xmlpdata;
        data^.text:=tmp;
        data^.hasSubordered:=false;
        isTag:=false;
        isValue:=false;
        isText:=false;
        if tmp='<' then isTag:=true else isText:=true;
      end;
    end;
    if isText and (tmp='<') then begin
      data^.text:=stringreplace(data^.text, #13, #10, [rfReplaceAll, rfIgnoreCase]);
      data^.text:=stringreplace(data^.text, #10#10, #10, [rfReplaceAll, rfIgnoreCase]);
      data^.hasSubordered:=false;
      //data^.text:=stringreplace(data^.text, #10, '', [rfReplaceAll, rfIgnoreCase]);
      new(data);
      preXMLList.Add(data);
      data^.ItemType:=xmlpTag;
      data^.text:=tmp;
      data^.hasSubordered:=false;
      isTag:=true;
      isValue:=false;
      isText:=false;
    end;

  end;
  CloseFile(F);



{  testoutput.clear;
  for i:=0 to preXMLList.count-1 do begin
    ID:='Tag   ';
    hs:='  no sub';
    if tjkpreXMLItem(preXMLList.items[i]^).hassubordered then hs:='     sub';
    if tjkpreXMLItem(preXMLList.items[i]^).ItemType=xmlpdata then ID:='Text  ';
    testOutput.add(ID+tjkpreXMLItem(preXMLList.items[i]^).text+hs);
  end;
  testoutput.add('');
  testoutput.add('');
  testoutput.add('');
  testoutput.add('');}



  Prolog.clear;
  XML.clear;
  err:=0;


  try
  prnt := Prolog; // zuerst wird der Prolog der Datei geparst
  isProlog:=true;
  ftags.clear;
  for i:=0 to preXMLList.Count-1 do begin
    data:=preXMLList[i];

    if data^.ItemType=xmlpdata then begin
      if (data^.text=']>')or(data^.text=']]>') then begin
        prnt:=prnt.ParentNode;
      end else if (not stringconsistsof(data^.text,[#10, #9, #32, #13]))and(data^.text<>'') then begin
        node:=prnt.AddChild;
        node.TagText:=data^.text;
        if FConvertSpaces then node.TagText:=convertXMLText(node.TagText);
        node.tagname:='';
        node.Nodetype:=xmlData;
        //testoutput.add('Text:...............'+ data^.text);
      end;
    end
    else if data^.ItemType=xmlptag then begin
      err:=InterpretTag(data^.text, tagname, tagtext, isend, isempty, sl, tagtype);
      if err=0 then begin
        if (isprolog)and(tagtype=xmlTag) then begin prnt:=XML; isProlog:=false; end;
        if not isend then begin
          node:=prnt.AddChild;
          node.TagText:=tagtext;
          node.tagname:=tagname;
          node.Nodetype:=tagtype;
          j:=0;
          if sl.count>0 then repeat
            try txt:=sl[j] except txt:='' end;
            node.ParamNames.Add(txt);
            inc(j);
            try txt:=sl[j] except txt:='' end;
            node.Paramvalues.Add(txt);
            inc(j);
          until j>=sl.count;
          if (TagType=xmlTag)and((not isempty)) then begin
            prnt:=node;
//            if ftags.count>0 then if ftags[ftags.count-1]=tagname then err:=7;
            ftags.add(tagname);
          end;
          if (TagType in [xmlDOCTYPE, xmlNameDecl])and(data^.hasSubordered) then prnt:=node;
          if (prnt=nil)or(not assigned(prnt)) then err:=5;
          if (err=0)and FTryInt then begin // versuchen bekannte Tags auszuwerten
            //xmlRoot, xmlTag, xmlData, xmlComment, xmlPI, xmlDOCTYPE, xmlATTLIST, xmlELEMENT, xmlENTITY, xmlNameDecl, xmlCDATA
            if TagType=xmlPI then begin
              if ansilowercase(tagname)='xml' then if InterpretPI_xml(data^.text)=0 then node.Parsed:=true;
            end;
            if TagType=xmlEntity then begin
              if InterpretEntityDef(data^.text)=0 then node.Parsed:=true;
            end;
          end;
         testoutput.add('Tag:................'+ tagname);
         if node.ParamNAMES.count>0 then testoutput.add('   Parameter 1:.....'+ node.ParamNames[0]+' = '+node.ParamValues[0]);
        end;
{        if isempty and(ftags.count>0) then
          if ftags[ftags.count-1]=tagname then err:=7;}
        if (TagType=xmlTag)and isend then begin
          prnt:=prnt.ParentNode;
          if (prnt=nil)or(not assigned(prnt)) then err:=5;
          if ftags.count>0 then begin
            if ftags[ftags.Count-1]=tagname then ftags.Delete(ftags.count-1) else begin err:=8 end;
          end else err:=8;

        end;
      end;
      if err>0 then begin if FShowError then showmessage(xmlError[err]+#13#10+'XML: '+data^.text); exit; end;
    end;

  end;
  sl.free;
  except if FShowError then showmessage(xmlError[err]+#13#10+'XML: '+data^.text); end;
  result:=err;
end;

function TjkXML.convertXMLText(text: string): string;
var i:longint;
    res:string;
    isText, isSpace:boolean;
begin
  result:='';
  isText:=false;
  isSpace:=true;
  res:='';
  for i:=1 to length(text) do begin
    if (res<>'')and(not(text[i] in xmlSpaces))then res:=res+text[i];
    if (res<>'')and(text[i] in xmlSpaces)then if(not(res[length(res)]in xmlSpaces))then res:=res+' ';
    if (res='')and(not(text[i] in xmlSpaces))then res:=res+text[i];
  end;
  for i:=length(res) downto 1 do begin
    if result<>'' then result:=res[i]+result;
    if (result='')and(not(res[i] in xmlSpaces)) then result:=res[i]+result;
  end;
end;

function TjkXMLNode.hasChildren: boolean;
begin
  result:=getChildCount>0; 
end;

function TjkXMLNode.getFirstChildNode(ntype: TjkxmlNodeType): TjkXMLNode;
var i:longint;
begin
  result:=nil;
  if hasCHildren then for i:=0 to getchildcount-1 do
    if child[i].Nodetype=ntype then begin result:=child[i]; break; end;
end;

function TjkXMLNode.getFirstChildNode(nam: string): TjkxmlNode;
var i:longint;
begin
  result:=nil;
  if hasCHildren then for i:=0 to getchildcount-1 do
    if child[i].TagName=nam then begin result:=child[i]; break; end;
end;

function TjkXMLNode.getParamValue(nam: string): string;
begin
  result:='';
  if ParamNames.IndexOf(nam)>=0 then
    result:=ParamValues[ParamNames.IndexOf(nam)];
end;

function TjkXMLNode.ParamExists(nam: string): boolean;
begin
  result:=ParamNames.IndexOf(nam)>=0;
end;

function TjkXMLNode.getParsedText: string;
begin
  result:=TagText;
  if (NodeType in [xmlData]) then result:=getParsedString(TagText);
end;

function TjkXMLNode.getParamValueParsed(nam: string): string;
begin
  result:='';
  if ParamNames.IndexOf(nam)>=0 then
    result:=getParsedString(ParamValues[ParamNames.IndexOf(nam)]);
end;

function TjkXML.ParseEntityString(txt: string): string;
var i:longint;
    isEntity, isPEntity :boolean;
    res, tmp:string;
    r:longint;
begin
  result:=Txt;
  if (length(txt)>2) then begin
    isEntity:=false;
    isPEntity:=false;
    res:='';
    tmp:='';
    for i:=1 to length(txt) do begin
      if txt[i]='&' then begin isEntity:=true; tmp:=''; end;
      //if txt[i]='%' then begin isPEntity:=true; tmp:=''; end;
      if not isEntity then res:=res+txt[i];
      if isEntity and (not(txt[i] in [';', '&', '%', ' ', #10, #13, #9])) then tmp:=tmp+txt[i];
      //if isPEntity and (not(txt[i] in [';', '&', '%', ' ', #10, #13, #9])) then tmp:=tmp+txt[i];
      //if isPEntity and (txt[i]=';') then begin isPEntity:=false; tmp:=''; end;
      if isEntity and (txt[i]=';') then begin
        isEntity:=false;
        if length(tmp)>0 then
          if copy(tmp,1,2)='#x' then try r:=hextodec(copy(tmp,3,length(tmp)-2)); res:=res+chr(r) except end
            else if tmp[1]='#' then try r:=strtoint(copy(tmp,2,length(tmp)-1)); res:=res+chr(r) except end
              else if tmp='apos' then res:=res+''''
                else if tmp='quot' then res:=res+'"'
                  else if tmp='amp' then res:=res+'&'
                    else if tmp='lt' then res:=res+'<'
                      else if tmp='gt' then res:=res+'>'
                        else if tmp='nbsp' then res:=res+' '//entspricht nicht ganz der XML-Definition !!!
                          else res:=res+getEntityData(tmp);

        tmp:='';
      end;
    end;
    //if isEntity then res:=res+tmp;
    result:=res;
  end;
end;

{ersetzt werden:

&  -->  &amp;
'  -->  &apos;
<  -->  &lt;
>  -->  &rt;
"  -->  &quot;

}
function EntityString(txt: string): string;
var i:longint;
    addtext:string;
begin
  result:='';
  if length(txt)>0 then begin
    for i:=1 to length(txt) do begin
      addtext:=txt[i];
      case txt[i] of
        '&': addtext:='&amp;';
        '''': addtext:='&apos;';
        '<': addtext:='&lt;';
        '>': addtext:='&rt;';
        '"': addtext:='&quot;';
      else
        if ord(txt[i]) in [0..31] then addtext:='&#'+inttostr(ord(txt[i]))+';';
      end;
      result:=result+addtext;
    end;
  end else result:=txt;
end;

end.
